@@ -3,10 +3,10 @@
  *  Module    : active.c
  *  Author    : I. Lea
  *  Created   : 1992-02-16
- *  Updated   : 2009-12-09
+ *  Updated   : 2016-06-02
  *  Notes     :
  *
- * Copyright (c) 1992-2010 Iain Lea <iain@bricbrac.de>
+ * Copyright (c) 1992-2016 Iain Lea <iain@bricbrac.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -43,8 +43,8 @@
 #endif /* !TCURSES_H */
 
 /*
- * List of allowed seperator chars in active file
- * unsed in parse_active_line()
+ * List of allowed separator chars in active file
+ * unused in parse_active_line()
  */
 #define ACTIVE_SEP	" \n"
 
@@ -66,15 +66,18 @@ static time_t active_timestamp;	/* time active file read (local) */
 static FILE *open_newgroups_fp(int idx);
 static FILE *open_news_active_fp(void);
 static int check_for_any_new_groups(void);
-static void active_add(struct t_group *ptr, long count, long max, long min, const char *moderated);
-static void append_group_line(char *active_file, char *group_path, long art_max, long art_min, char *base_dir);
+static void active_add(struct t_group *ptr, t_artnum count, t_artnum max, t_artnum min, const char *moderated);
+static void append_group_line(char *active_file, char *group_path, t_artnum art_max, t_artnum art_min, char *base_dir);
 static void make_group_list(char *active_file, char *base_dir, char *fixed_base, char *group_path);
 static void read_active_file(void);
 static void read_newsrc_active_file(void);
 static void subscribe_new_group(char *group, char *autosubscribe, char *autounsubscribe);
 #ifdef NNTP_ABLE
-	static t_bool parse_count_line(char *line, long *max, long *min, long *count, char *moderated);
+	static t_bool do_read_newsrc_active_file(FILE *fp);
+	static t_bool parse_count_line(char *line, t_artnum *max, t_artnum *min, t_artnum *count, char *moderated);
 	static void read_active_counts(void);
+#else
+	static void do_read_newsrc_active_file(FILE *fp);
 #endif /* NNTP_ABLE */
 
 
@@ -139,9 +142,9 @@ resync_active_file(
 static void
 active_add(
 	struct t_group *ptr,
-	long count,
-	long max,
-	long min,
+	t_artnum count,
+	t_artnum max,
+	t_artnum min,
 	const char *moderated)
 {
 	/* name - pre-initialised when group is made */
@@ -194,7 +197,7 @@ process_bogus(
 	if ((ptr = group_add(name)) == NULL)
 		return FALSE;
 
-	active_add(ptr, 0L, 1L, 0L, "n");
+	active_add(ptr, T_ARTNUM_CONST(0), T_ARTNUM_CONST(1), T_ARTNUM_CONST(0), "n");
 	ptr->bogus = TRUE;		/* Mark it bogus */
 
 	if (my_group_add(name, FALSE) < 0)
@@ -210,8 +213,8 @@ process_bogus(
 t_bool
 parse_active_line(
 	char *line,
-	long *max,
-	long *min,
+	t_artnum *max,
+	t_artnum *min,
 	char *moderated)
 {
 	char *p = NULL, *q = NULL, *r = NULL;
@@ -238,8 +241,8 @@ parse_active_line(
 		return FALSE;
 	}
 
-	*max = atol(p);
-	*min = atol(q);
+	*max = atoartnum(p);
+	*min = atoartnum(q);
 	strcpy(moderated, r);
 
 	return TRUE;
@@ -248,16 +251,15 @@ parse_active_line(
 
 #ifdef NNTP_ABLE
 /*
- * Parse line from "LIST COUNTS"
- * group high low count status, i.e.:
- * trigofacile.test 326 6 297 y
+ * Parse line from "LIST COUNTS" (RFC 6048)
+ * group high low count status
  */
 static t_bool
 parse_count_line(
 	char *line,
-	long *max,
-	long *min,
-	long *count,
+	t_artnum *max,
+	t_artnum *min,
+	t_artnum *count,
 	char *moderated)
 {
 	char *p = NULL, *q = NULL, *r = NULL, *s = NULL;
@@ -286,9 +288,9 @@ parse_count_line(
 		return FALSE;
 	}
 
-	*max = atol(p);
-	*min = atol(q);
-	*count = atol(r);
+	*max = atoartnum(p);
+	*min = atoartnum(q);
+	*count = atoartnum(r);
 	strcpy(moderated, s);
 
 	return TRUE;
@@ -304,19 +306,23 @@ parse_count_line(
  * We can't know the 'moderator' status and always return 'y'
  * But we don't change if the 'moderator' status is already checked by
  * read_active_file()
+ * Returnes TRUE if NNTP is enabled and authentication is needed
  */
+#ifdef NNTP_ABLE
+static t_bool
+#else
 static void
-read_newsrc_active_file(
-	void)
+#endif /* NNTP_ABLE */
+do_read_newsrc_active_file(
+	FILE *fp)
 {
-	FILE *fp;
 	char *ptr;
 	char *p;
 	char moderated[PATH_LEN];
 	int window = 0;
-	long count = -1L, min = 1L, max = 0L;
-	long processed = 0L;
-	static char ngname[NNTP_STRLEN];
+	t_artnum count = T_ARTNUM_CONST(-1), min = T_ARTNUM_CONST(1), max = T_ARTNUM_CONST(0);
+	t_artnum processed = T_ARTNUM_CONST(0);
+	static char ngname[NNTP_GRPLEN + 1]; /* RFC 3977 3.1 limits group names to 497 octets */
 	struct t_group *grpptr;
 #ifdef NNTP_ABLE
 	t_bool need_auth = FALSE;
@@ -325,17 +331,8 @@ read_newsrc_active_file(
 	int index_o = 0;
 #endif /* NNTP_ABLE */
 
-	/*
-	 * return immediately if no .newsrc can be found or .newsrc is empty
-	 * when function asked to use .newsrc
-	 */
-	if ((fp = fopen(newsrc, "r")) == NULL)
-		return;
 
-	if (file_size(newsrc) <= 0L) {
-		fclose(fp);
-		return;
-	}
+	rewind(fp);
 
 	if (!batch_mode || verbose)
 		wait_message(0, _(txt_reading_news_newsrc_file));
@@ -353,7 +350,7 @@ read_newsrc_active_file(
 			 * later on
 			 * TODO: check RFCs for possible max. size
 			 */
-			strncpy(ngname, ptr, 128);
+			my_strncpy(ngname, ptr, 128);
 			ptr = ngname;
 		}
 
@@ -361,7 +358,8 @@ read_newsrc_active_file(
 #ifdef NNTP_ABLE
 			char buf[NNTP_STRLEN];
 			char line[NNTP_STRLEN];
-			if (window < NUM_SIMULTANEOUS_GROUP_COMMAND && ptr) {
+
+			if (window < NUM_SIMULTANEOUS_GROUP_COMMAND && ptr && (!list_active || (newsrc_active && list_active && group_find(ptr, FALSE)))) {
 				ngnames[index_i] = my_strdup(ptr);
 				snprintf(buf, sizeof(buf), "GROUP %s", ngnames[index_i]);
 #	ifdef DEBUG
@@ -405,9 +403,9 @@ read_newsrc_active_file(
 
 					case OK_GROUP:
 						{
-							char fmt[20];
+							char fmt[25];
 
-							snprintf(fmt, sizeof(fmt), "%%ld %%ld %%ld %%%ds", NNTP_STRLEN);
+							snprintf(fmt, sizeof(fmt), "%%"T_ARTNUM_SFMT" %%"T_ARTNUM_SFMT" %%"T_ARTNUM_SFMT" %%%ds", NNTP_GRPLEN);
 							if (sscanf(line, fmt, &count, &min, &max, ngname) != 4) {
 								error_message(2, _(txt_error_invalid_response_to_group), line);
 #	ifdef DEBUG
@@ -419,7 +417,7 @@ read_newsrc_active_file(
 								error_message(2, _(txt_error_wrong_newsgroupname_in_group_response), ngname, ngnames[index_o], line);
 #	ifdef DEBUG
 								if (debug & DEBUG_NNTP) /* TODO: -> lang.c */
-									debug_print_file("NNTP", "Groupname missmatch in response to \"GROUP %s\": \"%s\"", ngnames[index_o], line);
+									debug_print_file("NNTP", "Groupname mismatch in response to \"GROUP %s\": \"%s\"", ngnames[index_o], line);
 #	endif /* DEBUG */
 							}
 							ptr = ngname;
@@ -501,10 +499,45 @@ read_newsrc_active_file(
 		 */
 		active_add(grpptr, count, max, min, moderated);
 	}
+#ifdef NNTP_ABLE
+	return need_auth;
+#endif /* NNTP_ABLE */
+}
+
+/*
+ * Wrapper for do_read_newsrc_active_file() to handle
+ * missing authentication
+ */
+static void
+read_newsrc_active_file(
+	void)
+{
+	FILE *fp;
+#ifdef NNTP_ABLE
+	t_bool need_auth;
+#endif /* NNTP_ABLE */
+
+	/*
+	 * return immediately if no .newsrc can be found or .newsrc is empty
+	 * when function asked to use .newsrc
+	 */
+	if ((fp = fopen(newsrc, "r")) == NULL)
+		return;
+
+	if (file_size(newsrc) <= 0L) {
+		fclose(fp);
+		return;
+	}
+
+#ifdef NNTP_ABLE
+	need_auth = do_read_newsrc_active_file(fp);
+#else
+	do_read_newsrc_active_file(fp);
+#endif /* NNTP_ABLE */
 
 #ifdef NNTP_ABLE
 	if (need_auth) { /* delayed auth */
-		if (!authenticate(nntp_server, userid, FALSE)) {
+		if (!authenticate(nntp_server, userid, FALSE) || do_read_newsrc_active_file(fp)) {
 			error_message(2, _(txt_auth_failed), ERR_ACCESS);
 			tin_done(EXIT_FAILURE);
 		}
@@ -539,9 +572,8 @@ open_news_active_fp(
 #ifdef NNTP_ABLE
 	if (read_news_via_nntp && !read_saved_news)
 		return (nntp_command("LIST", OK_GROUPS, NULL, 0));
-	else
 #endif /* NNTP_ABLE */
-		return (fopen(news_active_file, "r"));
+	return (fopen(news_active_file, "r"));
 }
 
 
@@ -555,7 +587,7 @@ read_active_file(
 	FILE *fp;
 	char *ptr;
 	char moderated[PATH_LEN];
-	long count = -1L, min = 1L, max = 0L;
+	t_artnum count = T_ARTNUM_CONST(-1), min = T_ARTNUM_CONST(1), max = T_ARTNUM_CONST(0);
 	long processed = 0L;
 	struct t_group *grpptr;
 
@@ -581,6 +613,11 @@ read_active_file(
 	}
 
 	while ((ptr = tin_fgets(fp, FALSE)) != NULL) {
+#if defined(DEBUG) && defined(NNTP_ABLE)
+		if (debug & DEBUG_NNTP)
+			debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
+#endif /* DEBUG && NNTP_ABLE */
+
 		if (!parse_active_line(ptr, &max, &min, moderated))
 			continue;
 
@@ -641,8 +678,8 @@ read_active_counts(
 	FILE *fp;
 	char *ptr;
 	char moderated[PATH_LEN];
-	long count = -1L, min = 1L, max = 0L;
-	long processed = 0L;
+	t_artnum count = T_ARTNUM_CONST(-1), min = T_ARTNUM_CONST(1), max = T_ARTNUM_CONST(0);
+	t_artnum processed = T_ARTNUM_CONST(0);
 	struct t_group *grpptr;
 
 	if (!batch_mode || verbose)
@@ -657,6 +694,11 @@ read_active_counts(
 	}
 
 	while ((ptr = tin_fgets(fp, FALSE)) != NULL) {
+#	ifdef DEBUG
+		if (debug & DEBUG_NNTP)
+			debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
+#	endif /* DEBUG */
+
 		if (!parse_count_line(ptr, &max, &min, &count, moderated))
 			continue;
 
@@ -702,7 +744,7 @@ read_active_counts(
 	if (!batch_mode || verbose)
 		my_fputc('\n', stdout);
 }
-#endif /* NNTP_ABLE*/
+#endif /* NNTP_ABLE */
 
 
 /*
@@ -716,6 +758,9 @@ read_news_active_file(
 	FILE *fp;
 	int newgrps = 0;
 	t_bool do_group_cmds = !nntp_caps.list_counts;
+#ifdef NNTP_ABLE
+	t_bool did_list_cmd = FALSE;
+#endif /* NNTP_ABLE */
 
 	/*
 	 * Ignore -n if no .newsrc can be found or .newsrc is empty
@@ -736,7 +781,8 @@ read_news_active_file(
 	/* Read an active file if it is allowed */
 	if (list_active) {
 #ifdef NNTP_ABLE
-		if (read_news_via_nntp && nntp_caps.type == CAPABILITIES && nntp_caps.list_counts)
+		did_list_cmd = TRUE;
+		if (read_news_via_nntp && nntp_caps.list_counts)
 			read_active_counts();
 		else
 #endif /* NNTP_ABLE */
@@ -748,23 +794,21 @@ read_news_active_file(
 #ifdef NNTP_ABLE
 #	ifndef DISABLE_PIPELINING
 		/*
-		 * use "LIST ACTIVE grp" (or even LIST ACTIVE grp,...) if we have
-		 * less than PIPELINE_LIMIT groups and we use -n but not -Q
-		 *
-		 * TODO: test me. do we want this overhead? add a DISABLE_PIPELINING
-		 *       code-path? we don't have list_active set but we use some
-		 *       sort of LIST ACTIVE -> our documentation is a bit incorrect
-		 *       now.
+		 * prefer LIST COUNTS, otherwise use LIST ACIVE (-l) or GROUP (-n)
+		 * or both (-ln); LIST COUNTS/ACTIVE grplist is used up to
+		 * PIPELINE_LIMIT groups in newsrc
 		 */
-		if (read_news_via_nntp && !list_active && ((nntp_caps.type == CAPABILITIES && nntp_caps.list_active) || nntp_caps.type != CAPABILITIES) && (show_description || check_for_new_newsgroups)) {
+		if (read_news_via_nntp && (list_active || nntp_caps.list_counts) && !did_list_cmd) {
 			char buff[NNTP_STRLEN];
 			char *ptr, *q;
 			char moderated[PATH_LEN];
-			int r = 0, i = 0, j = 0;
-			long count = -1L, min = 1L, max = 0L;
+			int r = 0, j = 0;
+			int i;
+			t_artnum count = T_ARTNUM_CONST(-1), min = T_ARTNUM_CONST(1), max = T_ARTNUM_CONST(0);
 			struct t_group *grpptr;
 			t_bool need_auth = FALSE;
 
+			*buff = '\0';
 			/* we can't use for_each_group(i) yet, so we have to prase the newsrc */
 			if ((fp = fopen(newsrc, "r")) != NULL) {
 				while (tin_fgets(fp, FALSE) != NULL)
@@ -777,21 +821,21 @@ read_news_active_file(
 						*q = '\0';
 						if (nntp_caps.type == CAPABILITIES && (nntp_caps.list_active || nntp_caps.list_counts)) {
 							/* LIST ACTIVE or LIST COUNTS takes wildmats */
-							if (!i) { /* new wildmat list */
-								i++;
-								snprintf(buff, sizeof(buff), "LIST %s %s", nntp_caps.list_counts ? "COUNTS" : "ACTIVE", ptr);
-								continue;
-							}
-							if (strlen(buff) + strlen(ptr) + 1 < NNTP_STRLEN) { /* append group name */
+							if (*buff && ((strlen(buff) + strlen(ptr)) < (NNTP_GRPLEN - 1))) { /* append group name */
 								snprintf(buff + strlen(buff), sizeof(buff) - strlen(buff), ",%s", ptr);
-								continue;
+							} else {
+								if (*buff) {
+									put_server(buff);
+									r++;
+								}
+								snprintf(buff, sizeof(buff), "LIST %s %s", nntp_caps.list_counts ? "COUNTS" : "ACTIVE", ptr);
 							}
-							i = 0; /* new wildmatlist required */
+							continue;
 						} else
 							snprintf(buff, sizeof(buff), "LIST ACTIVE %s", ptr);
 						put_server(buff);
 						r++;
-						*buff='\0';
+						*buff = '\0';
 					}
 					if (*buff) {
 						put_server(buff);
@@ -806,7 +850,7 @@ read_news_active_file(
 					for (i = 0; i < r && !did_reconnect; i++) {
 						if ((j = get_only_respcode(buff, sizeof(buff))) != OK_GROUPS) {
 							/* TODO: add 483 (RFC 3977) code */
-							if (j == ERR_NOAUTH || r == NEED_AUTHINFO)
+							if (j == ERR_NOAUTH || j == NEED_AUTHINFO)
 								need_auth = TRUE;
 #if 0 /* do we need something like this? */
 							if (j == ERR_CMDSYN)
@@ -817,7 +861,7 @@ read_news_active_file(
 							while ((ptr = tin_fgets(FAKE_NNTP_FP, FALSE)) != NULL) {
 #		ifdef DEBUG
 								if (debug & DEBUG_NNTP)
-									debug_print_file("NNTP", "<<< %s", ptr);
+									debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
 #		endif /* DEBUG */
 								if (nntp_caps.type == CAPABILITIES && nntp_caps.list_counts) {
 									if (!parse_count_line(ptr, &max, &min, &count, moderated))
@@ -881,8 +925,7 @@ read_news_active_file(
 
 /*
  * Open the active.times file locally or send the NEWGROUPS command
- *
- * NEWGROUPS yymmdd hhmmss
+ * "NEWGROUPS yymmdd hhmmss"
  */
 static FILE *
 open_newgroups_fp(
@@ -893,24 +936,26 @@ open_newgroups_fp(
 	struct tm *ngtm;
 
 	if (read_news_via_nntp && !read_saved_news) {
-		if (idx == -1)
+		/*
+		 * not checking for caps_type == CAPABILITIES && reader as some
+		 * servers do not support it even if advertizing READER so we must
+		 * handle errors anyway and just issue the cmd.
+		 */
+		if (idx == -1 || ((ngtm = localtime(&newnews[idx].time)) == NULL))
 			return (FILE *) 0;
 
-		ngtm = localtime(&newnews[idx].time);
 		/*
-		 * in the current draft, NEWGROUPS is allowed to take a 4 digit year
-		 * component - but even with a 2 digit year component it is y2k
-		 * compliant... we should switch over to ngtm->tm_year + 1900
-		 * when most servers can handle the new format
+		 * RFC 3077 states that we SHOULD use 4 digit year but some servers
+		 * still do not support it.
 		 */
 		snprintf(line, sizeof(line), "NEWGROUPS %02d%02d%02d %02d%02d%02d",
 			ngtm->tm_year % 100, ngtm->tm_mon + 1, ngtm->tm_mday,
 			ngtm->tm_hour, ngtm->tm_min, ngtm->tm_sec);
 
 		return (nntp_command(line, OK_NEWGROUPS, NULL, 0));
-	} else
+	}
 #endif /* NNTP_ABLE */
-		return (fopen(active_times_file, "r"));
+	return (fopen(active_times_file, "r"));
 }
 
 
@@ -1041,7 +1086,7 @@ subscribe_new_group(
 		}
 
 		if ((ptr = group_add(group)) != NULL)
-			active_add(ptr, 0L, 1L, 0L, "y");
+			active_add(ptr, T_ARTNUM_CONST(0), T_ARTNUM_CONST(1), T_ARTNUM_CONST(0), "y");
 
 		if ((idx = my_group_add(group, FALSE)) < 0)
 			return;
@@ -1202,9 +1247,6 @@ find_newnews_index(
 /*
  * Get a single status char from the moderated field. Used on selection screen
  * and in header of group screen
- *
- * TODO: what about 'j' groups? active(5) says:
- *       "local postings to that group should not be generated"
  */
 char
 group_flag(
@@ -1216,6 +1258,7 @@ group_flag(
 
 		case 'x':
 		case 'n':
+		case 'j':
 			return 'X';
 
 		case '=':
@@ -1241,15 +1284,17 @@ create_save_active_file(
 	if (no_write && file_size(local_save_active_file) != -1L)
 		return;
 
-	wait_message(0, _(txt_creating_active));
+	if (strfpath(cmdline.args & CMDLINE_SAVEDIR ? cmdline.savedir : tinrc.savedir, group_path, sizeof(group_path), NULL, FALSE)) {
+		wait_message(0, _(txt_creating_active));
+		print_active_head(local_save_active_file);
 
-	print_active_head(local_save_active_file);
-	strfpath(cmdline.args & CMDLINE_SAVEDIR ? cmdline.savedir : tinrc.savedir, group_path, sizeof(group_path), NULL, FALSE);
-	while (strlen(group_path) && group_path[strlen(group_path) - 1] == '/')
-		group_path[strlen(group_path) - 1] = '\0';
-	fb = my_strdup(group_path);
-	make_group_list(local_save_active_file, cmdline.args & CMDLINE_SAVEDIR ? cmdline.savedir : tinrc.savedir, fb, group_path);
-	free(fb);
+		while (strlen(group_path) && group_path[strlen(group_path) - 1] == '/')
+			group_path[strlen(group_path) - 1] = '\0';
+
+		fb = my_strdup(group_path);
+		make_group_list(local_save_active_file, cmdline.args & CMDLINE_SAVEDIR ? cmdline.savedir : tinrc.savedir, fb, group_path);
+		free(fb);
+	}
 }
 
 
@@ -1265,8 +1310,8 @@ make_group_list(
 	char *ptr;
 	char filename[PATH_LEN];
 	char path[PATH_LEN];
-	long art_max;
-	long art_min;
+	t_artnum art_max;
+	t_artnum art_min;
 	struct stat stat_info;
 	t_bool is_dir;
 
@@ -1302,8 +1347,8 @@ static void
 append_group_line(
 	char *active_file,
 	char *group_path,
-	long art_max,
-	long art_min,
+	t_artnum art_max,
+	t_artnum art_min,
 	char *base_dir)
 {
 	FILE *fp;
@@ -1329,14 +1374,18 @@ append_group_line(
 		while ((ptr = strchr(ptr, '/')) != NULL)
 			*ptr = '.';
 
-		wait_message(0, "Appending=[%s %ld %ld %s]\n", group_name, art_max, art_min, base_dir);
+		wait_message(0, "Appending=[%s %"T_ARTNUM_PFMT" %"T_ARTNUM_PFMT" %s]\n", group_name, art_max, art_min, base_dir);
 		print_group_line(fp, group_name, art_max, art_min, base_dir);
 		if ((err = ferror(fp)) || fclose(fp)) { /* TODO: issue warning? */
-			rename(file_tmp, active_file);
 			if (err) {
 				clearerr(fp);
 				fclose(fp);
 			}
+			err = rename(file_tmp, active_file);
+#ifdef DEBUG
+			if ((debug & DEBUG_MISC) && err) /* TODO: is this the right debug-level? */
+				perror_message(_(txt_rename_error), file_tmp, active_file);
+#endif /* DEBUG */
 		}
 		free(group_name);
 	}
