@@ -3,11 +3,11 @@
  *  Module    : auth.c
  *  Author    : Dirk Nimmich <nimmich@muenster.de>
  *  Created   : 1997-04-05
- *  Updated   : 2009-06-27
+ *  Updated   : 2016-01-31
  *  Notes     : Routines to authenticate to a news server via NNTP.
  *              DON'T USE get_respcode() THROUGHOUT THIS CODE.
  *
- * Copyright (c) 1997-2010 Dirk Nimmich <nimmich@muenster.de>
+ * Copyright (c) 1997-2016 Dirk Nimmich <nimmich@muenster.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -61,8 +61,8 @@ static t_bool authinfo_plain(char *server, char *authuser, t_bool startup);
 
 
 /*
- * Read the $HOME/.newsauth file and put authentication username
- * and password for the specified server in the given strings.
+ * Read the ${TIN_HOMEDIR:-"$HOME"}/.newsauth file and put authentication
+ * username and password for the specified server in the given strings.
  * Returns TRUE if at least a password was found, FALSE if there was
  * no .newsauth file or no matching server.
  */
@@ -105,7 +105,7 @@ read_newsauth_file(
 
 		/*
 		 * Search through authorization file for correct NNTP server
-		 * File has format:  'nntp-server' 'password' ['username']
+		 * File has format: 'nntp-server' 'password' ['username']
 		 */
 		while (fgets(line, sizeof(line), fp) != NULL) {
 			/* strip trailing newline character */
@@ -154,8 +154,8 @@ read_newsauth_file(
 		}
 		fclose(fp);
 		return (found > 0);
-	} else
-		return FALSE;
+	}
+	return FALSE;
 }
 
 
@@ -235,9 +235,8 @@ authinfo_plain(
 	static t_bool already_failed = FALSE;
 	static t_bool initialized = FALSE;
 
-	changed = strcmp(server, last_server);	/* do we need new auth values? */
-	strncpy(last_server, server, PATH_LEN - 1);
-	last_server[PATH_LEN - 1] = '\0';
+	if ((changed = strcmp(server, last_server)))	/* do we need new auth values? */
+		STRCPY(last_server, server);
 
 	/*
 	 * Let's try the previous auth pair first, if applicable.
@@ -250,14 +249,15 @@ authinfo_plain(
 		if (ret != OK_AUTH)
 #	endif /* USE_SASL */
 		{
-			if (nntp_caps.type != CAPABILITIES || (nntp_caps.type == CAPABILITIES && nntp_caps.authinfo_user))
+			if (nntp_caps.type != CAPABILITIES || nntp_caps.authinfo_user)
 				ret = do_authinfo_user(server, authusername, authpassword);
 		}
 		return (ret == OK_AUTH);
 	}
 
 	authpassword[0] = '\0';
-	authuser = strncpy(authusername, authuser, sizeof(authusername) - 1);
+	STRCPY(authusername, authuser);
+	authuser = authusername;
 	authpass = authpassword;
 
 	/*
@@ -304,15 +304,19 @@ authinfo_plain(
 	 * .newsauth has failed or there's no .newsauth file respectively no
 	 * matching username/password for the current server. If we are not at
 	 * startup we ask the user to enter such a pair by hand. Don't ask him
-	 * startup except if requested by -A option because if he doesn't need
-	 * authenticate(we don't know), the "Server expects authentication"
+	 * at startup except if requested by -A option because if he doesn't need
+	 * to authenticate (we don't know), the "Server expects authentication"
 	 * messages are annoying (and even wrong).
 	 * UNSURE: Maybe we want to make this decision configurable in the
 	 * options menu, too, so that the user doesn't need -A.
 	 * TODO: Put questions into do_authinfo_user() because it is possible
 	 * that the server doesn't want a password; so only ask for it if needed.
 	 */
-	 if (force_auth_on_conn_open || !startup) {
+	if (force_auth_on_conn_open || !startup) {
+		if (batch_mode) { /* no interactive username/password prompting */
+			error_message(0, _(txt_auth_needed));
+			return (ret == OK_AUTH);
+		}
 		if (nntp_caps.type != CAPABILITIES || (nntp_caps.type == CAPABILITIES && !nntp_caps.authinfo_state && ((nntp_caps.sasl & SASL_PLAIN) || nntp_caps.authinfo_user || (!nntp_caps.authinfo_user && !(nntp_caps.sasl & SASL_PLAIN))))) {
 #	ifdef USE_CURSES
 			int state = RawState();
@@ -322,7 +326,7 @@ authinfo_plain(
 #	ifdef USE_CURSES
 			Raw(TRUE);
 #	endif /* USE_CURSES */
-			if (!prompt_default_string(_(txt_auth_user), authuser, PATH_LEN, authusername, HIST_NONE)) {
+			if (!prompt_default_string(_(txt_auth_user), authuser, sizeof(authusername) - 1, authusername, HIST_NONE)) {
 #	ifdef DEBUG
 				if (debug & DEBUG_NNTP)
 					debug_print_file("NNTP", "authorization failed: no username");
@@ -333,14 +337,15 @@ authinfo_plain(
 #	ifdef USE_CURSES
 			Raw(state);
 			my_printf("%s", _(txt_auth_pass));
-			wgetnstr(stdscr, authpassword, sizeof(authpassword));
+			wgetnstr(stdscr, authpassword, sizeof(authpassword) - 1);
+			authpassword[sizeof(authpassword) - 1] = '\0';
 			Raw(TRUE);
 #	else
 			/*
-			 * on some systems (i.e. Solaris) getpass(3) is limited to 8 chars ->
-			 * we use tin_getline()
+			 * on some systems (i.e. Solaris) getpass(3) is limited
+			 * to 8 chars -> we use tin_getline()
 			 */
-			authpass = strncpy(authpassword, tin_getline(_(txt_auth_pass), FALSE, NULL, PATH_LEN, TRUE, HIST_NONE), sizeof(authpassword) - 1);
+			STRCPY(authpassword, tin_getline(_(txt_auth_pass), FALSE, NULL, sizeof(authpassword) - 1, TRUE, HIST_NONE));
 #	endif /* USE_CURSES */
 
 #	ifdef USE_SASL
@@ -349,7 +354,7 @@ authinfo_plain(
 			if (ret != OK_AUTH)
 #	endif /* USE_SASL */
 			{
-				if (nntp_caps.type != CAPABILITIES || (nntp_caps.type == CAPABILITIES && (nntp_caps.authinfo_user || (!nntp_caps.authinfo_user && !nntp_caps.authinfo_sasl)))) {
+				if (nntp_caps.type != CAPABILITIES || (nntp_caps.authinfo_user || !nntp_caps.authinfo_sasl)) {
 #	ifdef DEBUG
 					if (debug & DEBUG_NNTP) {
 						if (nntp_caps.type == CAPABILITIES && !nntp_caps.authinfo_sasl && !nntp_caps.authinfo_user)
@@ -386,7 +391,7 @@ authinfo_plain(
 #	endif /* DEBUG */
 			/*
 			 * we return OK_AUTH here once so tin doesn't exit just because a
-			 * single comand requested auth ...
+			 * single command requested auth ...
 			 */
 			if (!already_failed)
 				ret = OK_AUTH;
@@ -427,18 +432,70 @@ do_authinfo_sasl_plain(
 {
 	char line[PATH_LEN];
 	char *foo;
+	char *utf8user;
+	char *utf8pass;
 	int ret;
+#		ifdef CHARSET_CONVERSION
+	char *cp;
+	int i, c = 0;
+	t_bool contains_8bit = FALSE;
+#		endif /* CHARSET_CONVERSION */
+
+	utf8user = my_strdup(authuser);
+	utf8pass = my_strdup(authpass);
+#		ifdef CHARSET_CONVERSION
+	/* RFC 4616 */
+	if (!IS_LOCAL_CHARSET("UTF-8")) {
+		for (cp = utf8user; *cp && !contains_8bit; cp++) {
+			if (!isascii(*cp)) {
+				contains_8bit = TRUE;
+				break;
+			}
+		}
+		for (cp = utf8pass; *cp && !contains_8bit; cp++) {
+			if (!isascii(*cp)) {
+				contains_8bit = TRUE;
+				break;
+			}
+		}
+		if (contains_8bit) {
+			for (i = 0; txt_mime_charsets[i] != NULL; i++) {
+				if (!strcasecmp("UTF-8", txt_mime_charsets[i])) {
+					c = i;
+					break;
+				}
+			}
+			if (c == i) { /* should never fail */
+				if (!buffer_to_network(utf8user, c)) {
+					free(utf8user);
+					utf8user = my_strdup(authuser);
+				} else {
+					if (!buffer_to_network(utf8pass, c)) {
+						free(utf8pass);
+						utf8pass = my_strdup(authpass);
+					}
+				}
+			}
+		}
+	}
+#		endif /* CHARSET_CONVERSION */
 
 #		ifdef DEBUG
 	if (debug & DEBUG_NNTP)
-		debug_print_file("NNTP", "do_authinfo_sasl_plain(%s, %s)", BlankIfNull(authuser), BlankIfNull(authpass));
+		debug_print_file("NNTP", "do_authinfo_sasl_plain(\"%s\", \"%s\")", BlankIfNull(authuser), BlankIfNull(authpass));
 #		endif /* DEBUG */
 
-	if ((foo = sasl_auth_plain(authuser, authpass)) == NULL)
+	if ((foo = sasl_auth_plain(utf8user, utf8pass)) == NULL) {
+		free(utf8user);
+		free(utf8pass);
 		return ERR_AUTHBAD;
+	}
+
+	free(utf8user);
+	free(utf8pass);
 
 	snprintf(line, sizeof(line), "AUTHINFO SASL PLAIN %s", foo);
-	free(foo);
+	FreeIfNeeded(foo);
 #		ifdef DEBUG
 	if (debug & DEBUG_NNTP)
 		debug_print_file("NNTP", "authorization %s", line);
@@ -451,7 +508,8 @@ do_authinfo_sasl_plain(
 }
 
 
-static char *sasl_auth_plain(
+static char *
+sasl_auth_plain(
 	char *user,
 	char *pass)
 {
