@@ -3,10 +3,10 @@
  *  Module    : main.c
  *  Author    : I. Lea & R. Skrenta
  *  Created   : 1991-04-01
- *  Updated   : 2009-12-09
+ *  Updated   : 2016-02-02
  *  Notes     :
  *
- * Copyright (c) 1991-2010 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
+ * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -77,9 +77,8 @@ main(
 	int argc,
 	char *argv[])
 {
-	int count;
+	int count, start_groupnum;
 	int num_cmd_line_groups = 0;
-	int start_groupnum = 0;
 	t_bool tmp_no_write;
 
 	cmd_line = TRUE;
@@ -88,8 +87,8 @@ main(
 #if defined(HAVE_SETLOCALE) && !defined(NO_LOCALE)
 	if (setlocale(LC_ALL, "")) {
 #	ifdef ENABLE_NLS
-		bindtextdomain(PACKAGE, LOCALEDIR);
-		textdomain(PACKAGE);
+		bindtextdomain(NLS_TEXTDOMAIN, LOCALEDIR);
+		textdomain(NLS_TEXTDOMAIN);
 #	endif /* ENABLE_NLS */
 	} else
 		error_message(4, txt_error_locale);
@@ -118,6 +117,7 @@ main(
 
 	debug = 0;	/* debug OFF */
 
+	tin_progname = my_malloc(strlen(argv[0]) + 1);
 	base_name(argv[0], tin_progname);
 
 #ifdef NNTP_ONLY
@@ -131,6 +131,7 @@ main(
 		read_news_via_nntp = TRUE;
 #	else
 		error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+		free(tin_progname);
 		giveup();
 #	endif /* NNTP_ABLE */
 	}
@@ -143,7 +144,6 @@ main(
 	hash_init();
 	init_selfinfo();
 	init_group_hash();
-	setup_default_keys(); /* preinit keybindings */
 
 	/*
 	 * Process envargs & command line options
@@ -151,6 +151,10 @@ main(
 	 */
 	read_cmd_line_options(argc, argv);
 
+	/* preinit keybindings if interactive */
+	if (!batch_mode)
+		setup_default_keys();
+
 	/*
 	 * Read user local & global config files
 	 * These override the compiled in defaults
@@ -168,6 +172,7 @@ main(
 #	ifndef USE_CURSES
 		if (!get_termcaps()) {
 			error_message(2, _(txt_screen_init_failed), tin_progname);
+			free_all_arrays();
 			giveup();
 		}
 #	endif /* !USE_CURSES */
@@ -178,6 +183,7 @@ main(
 		 */
 		if (!InitScreen()) {
 			error_message(2, _(txt_screen_init_failed), tin_progname);
+			free_all_arrays();
 			giveup();
 		}
 
@@ -195,9 +201,13 @@ main(
 	/*
 	 * Connect to nntp server?
 	 */
-	nntp_server = getserverbyfile(NNTP_SERVER_FILE);
-	if (read_news_via_nntp && !read_saved_news && nntp_open())
+	if (!nntp_server || !*nntp_server)
+		nntp_server = getserverbyfile(NNTP_SERVER_FILE);
+	if (read_news_via_nntp && !read_saved_news && nntp_open()) {
+		nntp_close();
+		free_all_arrays();
 		giveup();
+	}
 
 	read_server_config();
 
@@ -210,7 +220,8 @@ main(
 	 * no error message? why?
 	 */
 	if (update_index && nntp_caps.over_cmd && !tinrc.cache_overview_files) {
-		error_message(2, _(txt_batch_update_unavail), tin_progname);
+		error_message(2, _(txt_batch_update_unavail), tin_progname, print_boolean(tinrc.cache_overview_files));
+		free_all_arrays();
 		giveup();
 	}
 
@@ -275,7 +286,7 @@ main(
 	 * Read in users filter preferences file. This has to be done before
 	 * quick post because the filters might be updated.
 	 */
-	filtered_articles = read_filter_file(filter_file);
+	read_filter_file(filter_file);
 
 	no_write = TRUE;
 #ifdef DEBUG
@@ -352,8 +363,10 @@ main(
 
 	if (start_any_unread) {
 		batch_mode = TRUE;			/* Suppress some unwanted on-screen garbage */
-		if ((start_groupnum = check_start_save_any_news(START_ANY_NEWS, catchup)) == -1)
-			giveup();				/* No new/unread news so exit */
+		if ((start_groupnum = check_start_save_any_news(START_ANY_NEWS, catchup)) == -1) {
+			batch_mode = FALSE;
+			tin_done(EXIT_SUCCESS);
+		}
 		batch_mode = FALSE;
 	}
 
@@ -416,8 +429,11 @@ main(
 
 /*
  * process command line options
+ * [01235789beEFijJkKLOtTyY] are unused
+ * [W] is reserved
+ * [BCPU] have been in use at some time, but now are unused
  */
-#define OPTIONS "aAcdD:f:g:G:hHI:lm:M:nNop:qQrRs:SuvVwxXzZ"
+#define OPTIONS "46aAcdD:f:g:G:hHI:lm:M:nNop:qQrRs:SuvVwxXzZ"
 
 static void
 read_cmd_line_options(
@@ -431,11 +447,47 @@ read_cmd_line_options(
 
 	while ((ch = getopt(argc, argv, OPTIONS)) != -1) {
 		switch (ch) {
+
+			case '4':
+#if defined(NNTP_ABLE) && defined(INET6)
+				force_ipv4 = TRUE;
+				read_news_via_nntp = TRUE;
+#else
+#	ifdef NNTP_ABLE
+				error_message(2, _(txt_option_not_enabled), "-DENABLE_IPV6");
+#	else
+				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+#	endif /* NNTP_ABLE */
+				free_all_arrays();
+				giveup();
+				/* keep lint quiet: */
+				/* NOTREACHED */
+#endif /* NNTP_ABLE && INET6 */
+				break;
+
+			case '6':
+#if defined(NNTP_ABLE) && defined(INET6)
+				force_ipv6 = TRUE;
+				read_news_via_nntp = TRUE;
+#	else
+#	ifdef NNTP_ABLE
+				error_message(2, _(txt_option_not_enabled), "-DENABLE_IPV6");
+#	else
+				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+#	endif /* NNTP_ABLE */
+				free_all_arrays();
+				giveup();
+				/* keep lint quiet: */
+				/* NOTREACHED */
+#endif /* NNTP_ABLE && INET6 */
+				break;
+
 			case 'a':
 #ifdef HAVE_COLOR
 				cmdline.args |= CMDLINE_USE_COLOR;
 #else
 				error_message(2, _(txt_option_not_enabled), "-DHAVE_COLOR");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -445,8 +497,10 @@ read_cmd_line_options(
 			case 'A':
 #ifdef NNTP_ABLE
 				force_auth_on_conn_open = TRUE;
+				read_news_via_nntp = TRUE;
 #else
 				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -468,6 +522,7 @@ read_cmd_line_options(
 				debug_delete_files();
 #else
 				error_message(2, _(txt_option_not_enabled), "-DDEBUG");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -491,6 +546,7 @@ read_cmd_line_options(
 				read_news_via_nntp = TRUE;
 #else
 				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -499,6 +555,7 @@ read_cmd_line_options(
 
 			case 'H':
 				show_intro_page();
+				free_all_arrays();
 				exit(EXIT_SUCCESS);
 				/* keep lint quiet: */
 				/* FALLTHROUGH */
@@ -542,6 +599,7 @@ read_cmd_line_options(
 				check_for_new_newsgroups = FALSE;
 #else
 				error_message(2, _(txt_option_not_enabled), "-UNO_POSTING");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -555,6 +613,7 @@ read_cmd_line_options(
 					nntp_tcp_port = (unsigned short) atoi(optarg);
 #else
 				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -576,6 +635,7 @@ read_cmd_line_options(
 				read_news_via_nntp = TRUE;
 #else
 				error_message(2, _(txt_option_not_enabled), "-DNNTP_ABLE");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -611,6 +671,7 @@ read_cmd_line_options(
 
 			case 'V':
 				tin_version_info(stderr);
+				free_all_arrays();
 				exit(EXIT_SUCCESS);
 				/* keep lint quiet: */
 				/* FALLTHROUGH */
@@ -621,6 +682,7 @@ read_cmd_line_options(
 				check_for_new_newsgroups = FALSE;
 #else
 				error_message(2, _(txt_option_not_enabled), "-UNO_POSTING");
+				free_all_arrays();
 				giveup();
 				/* keep lint quiet: */
 				/* NOTREACHED */
@@ -653,6 +715,7 @@ read_cmd_line_options(
 			case '?':
 			default:
 				usage(tin_progname);
+				free_all_arrays();
 				exit(EXIT_SUCCESS);
 		}
 	}
@@ -660,9 +723,10 @@ read_cmd_line_options(
 	num_cmdargs = optind;
 	max_cmdargs = argc;
 	if (!newsrc_set) {
-		if (read_news_via_nntp)
-			get_newsrcname(newsrc, sizeof(newsrc), getserverbyfile(NNTP_SERVER_FILE));
-		else {
+		if (read_news_via_nntp) {
+			nntp_server = getserverbyfile(NNTP_SERVER_FILE);
+			get_newsrcname(newsrc, sizeof(newsrc), nntp_server);
+		} else {
 #if defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
 			struct utsname uts;
 			(void) uname(&uts);
@@ -731,6 +795,21 @@ read_cmd_line_options(
 		wait_message(2, _(txt_useless_combination), "-Z", "-z", "-Z");
 		check_any_unread = FALSE;
 	}
+#	ifdef DEBUG
+	if ((debug & DEBUG_NNTP) && !read_news_via_nntp) {
+		/* TODO: lang.c */
+		wait_message(3, _(txt_useless_combination), _("reading from local spool"), "-D nntp", "-D nntp");
+		debug &= ~DEBUG_NNTP;
+	}
+#	endif /* DEBUG */
+
+#if defined(NNTP_ABLE) && defined(INET6)
+	if (force_ipv4 && force_ipv6) {
+		wait_message(2, _(txt_useless_combination), "-4", "-6", "-6");
+		force_ipv6 = FALSE;
+	}
+#endif /* NNTP_ABLE && INET6 */
+
 	if (mail_news || save_news || update_index || check_any_unread || catchup)
 		batch_mode = TRUE;
 	else
@@ -773,6 +852,11 @@ usage(
 {
 	error_message(2, _(txt_usage_tin), theProgname);
 
+#if defined(NNTP_ABLE) && defined(INET6)
+		error_message(2, _(txt_usage_force_ipv4));
+		error_message(2, _(txt_usage_force_ipv6));
+#endif /* NNTP_ABLE && INET6 */
+
 #ifdef HAVE_COLOR
 		error_message(2, _(txt_usage_toggle_color));
 #endif /* HAVE_COLOR */
@@ -901,12 +985,11 @@ read_cmd_line_groups(
 			for_each_group(i) {
 				if (match_group_list(active[i].name, cmdargs[num])) {
 					if (my_group_add(active[i].name, TRUE) != -1) {
+						matched++;
 						if (post_article_and_exit) {
 							my_strncpy(tinrc.default_post_newsgroups, active[i].name, sizeof(tinrc.default_post_newsgroups) - 1);
-							matched++;
 							break;
 						}
-						matched++;
 					}
 				}
 			}
@@ -945,6 +1028,7 @@ create_mail_save_dirs(
  * giveup() indicates a serious error and keeping track of what we've
  * already malloc()ed would be a PITA.
  */
+/* coverity[+kill] */
 void
 giveup(
 	void)
