@@ -3,10 +3,10 @@
  *  Module    : select.c
  *  Author    : I. Lea & R. Skrenta
  *  Created   : 1991-04-01
- *  Updated   : 2009-10-30
+ *  Updated   : 2016-06-06
  *  Notes     :
  *
- * Copyright (c) 1991-2010 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
+ * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -43,6 +43,8 @@
 #endif /* !TCURSES_H */
 
 
+static struct t_fmt sel_fmt;
+
 /*
  * Local prototypes
  */
@@ -64,6 +66,11 @@ static void sort_active_file(void);
 static void subscribe_pattern(const char *prompt, const char *message, const char *result, t_bool state);
 static void sync_active_file(void);
 static void yank_active_file(void);
+#ifdef NNTP_ABLE
+	static char *lookup_msgid(char *id);
+	static int show_article_by_msgid(void);
+	static struct t_group *get_group_from_list(char *newsgroups);
+#endif /* NNTP_ABLE */
 
 
 /*
@@ -75,6 +82,8 @@ static void yank_active_file(void);
 t_menu selmenu = { 1, 0, 0, show_selection_page, draw_group_arrow, build_gline };
 
 static int groupname_len;	/* max. group name length */
+static int flags_offset;
+static int ucnt_offset;
 
 
 static t_function
@@ -109,6 +118,7 @@ selection_page(
 	setbuf(stdin, 0);
 #endif /* READ_CHAR_HACK */
 
+	Raw(TRUE);
 	ClearScreen();
 
 	/*
@@ -205,7 +215,7 @@ selection_page(
 			case GLOBAL_SEARCH_SUBJECT_BACKWARD:
 			case GLOBAL_SEARCH_REPEAT:
 				if (func == GLOBAL_SEARCH_REPEAT && last_search != GLOBAL_SEARCH_SUBJECT_FORWARD && last_search != GLOBAL_SEARCH_SUBJECT_BACKWARD)
-					info_message(_(txt_bad_command), printascii(key, key_to_func(GLOBAL_HELP, select_keys)));
+					info_message(_(txt_no_prev_search));
 				else {
 					if ((i = search_active((func == GLOBAL_SEARCH_SUBJECT_FORWARD), (func == GLOBAL_SEARCH_REPEAT))) != -1) {
 						move_to_item(i);
@@ -251,16 +261,19 @@ selection_page(
 				break;
 
 			case GLOBAL_EDIT_FILTER:
-				if (!invoke_editor(filter_file, filter_file_offset, NULL))
-					break;
-				(void) read_filter_file(filter_file);
+				if (invoke_editor(filter_file, filter_file_offset, NULL))
+					(void) read_filter_file(filter_file);
+				show_selection_page();
 				break;
 
 			case SELECT_TOGGLE_DESCRIPTIONS:	/* toggle newsgroup descriptions */
-				show_description = bool_not(show_description);
-				if (show_description)
-					read_descriptions(TRUE);
-				show_selection_page();
+				if (sel_fmt.show_grpdesc) {
+					show_description = bool_not(show_description);
+					if (show_description)
+						read_descriptions(TRUE);
+					show_selection_page();
+				} else
+					info_message(_(txt_grpdesc_disabled));
 				break;
 
 			case SELECT_GOTO:			/* prompt for a new group name */
@@ -284,6 +297,23 @@ selection_page(
 				show_selection_page();
 				break;
 
+#ifdef NNTP_ABLE
+			case GLOBAL_LOOKUP_MESSAGEID:
+				switch (n = show_article_by_msgid()) {
+					case 0:
+						show_selection_page();
+						break;
+
+					case GRP_QUIT:
+						select_quit();
+						break;
+
+					default:
+						break;
+				}
+				break;
+#endif /* NNTP_ABLE */
+
 			case GLOBAL_TOGGLE_HELP_DISPLAY:	/* toggle mini help menu */
 				toggle_mini_help(SELECT_LEVEL);
 				show_selection_page();
@@ -342,7 +372,7 @@ selection_page(
 				break;
 
 			case GLOBAL_OPTION_MENU:
-				config_page(CURR_GROUP.name);
+				config_page(selmenu.max ? CURR_GROUP.name : NULL, signal_context);
 				show_selection_page();
 				break;
 
@@ -381,6 +411,8 @@ selection_page(
 				show_selection_page();
 				if (tinrc.show_only_unread_groups)
 					info_message(_(txt_show_unread));
+				else
+					clear_message();
 				break;
 
 			case GLOBAL_BUGREPORT:
@@ -422,7 +454,7 @@ selection_page(
 					break;
 				}
 				if (CURR_GROUP.subscribed) {
-					mark_screen(selmenu.curr, 2, CURR_GROUP.newgroup ? "N" : "u");
+					mark_screen(selmenu.curr, flags_offset, CURR_GROUP.newgroup ? "N" : "u");
 					subscribe(&CURR_GROUP, UNSUBSCRIBED, TRUE);
 					info_message(_(txt_unsubscribed_to), CURR_GROUP.name);
 					move_down();
@@ -470,7 +502,7 @@ selection_page(
 						/*
 						 * this is a gross hack to avoid a crash in the
 						 * CHARSET_CONVERSION conversion case in new_part()
-						 * which relies currently relies on CURR_GROUP
+						 * which currently relies on CURR_GROUP
 						 */
 						selmenu.curr = my_group_add(buf, FALSE);
 						/*
@@ -482,7 +514,7 @@ selection_page(
 #endif /* 1 */
 					}
 				} else
-					strcpy(buf, CURR_GROUP.name);
+					STRCPY(buf, CURR_GROUP.name);
 				if (!can_post && !CURR_GROUP.bogus && !CURR_GROUP.attribute->mailing_list) {
 					info_message(_(txt_cannot_post));
 					break;
@@ -520,15 +552,16 @@ selection_page(
 					break;
 				}
 				grp_mark_unread(&CURR_GROUP);
-				{
-					char tmp[6];
+				if (CURR_GROUP.newsrc.num_unread)
+					STRCPY(buf, tin_ltoa(CURR_GROUP.newsrc.num_unread, sel_fmt.len_ucnt));
+				else {
+					size_t j = 0;
 
-					if (CURR_GROUP.newsrc.num_unread)
-						STRCPY(tmp, tin_ltoa(CURR_GROUP.newsrc.num_unread, 5));
-					else
-						STRCPY(tmp, "     ");
-					mark_screen(selmenu.curr, 9, tmp);
+					while (j < sel_fmt.len_ucnt)
+						buf[j++] = ' ';
+					buf[j] = '\0';
 				}
+				mark_screen(selmenu.curr, ucnt_offset, buf);
 				break;
 
 			default:
@@ -543,10 +576,16 @@ show_selection_page(
 	void)
 {
 	char buf[LEN];
-	int i, len;
+	int i;
+	size_t len;
 
 	signal_context = cSelect;
 	currmenu = &selmenu;
+	parse_format_string(tinrc.select_format, &sel_fmt);
+	groupname_len = 0;
+	flags_offset = 0;
+	mark_offset = 0;
+	ucnt_offset = 0;
 
 	if (read_news_via_nntp)
 		snprintf(buf, sizeof(buf), "%s (%s  %d%s)", _(txt_group_selection), nntp_server, selmenu.max, (tinrc.show_only_unread_groups ? _(" R") : ""));
@@ -560,37 +599,42 @@ show_selection_page(
 	set_first_screen_item();
 	show_title(buf);
 
-	/*
-	 * calculate max length of groupname field
-	 * if yanked in (yanked_out == FALSE) check all groups in active file
-	 * otherwise just subscribed to groups
-	 */
-	if (yanked_out) {
-		for (i = 0; i < selmenu.max; i++) {
-			if ((len = strlen(active[my_group[i]].name)) > groupname_len)
-				groupname_len = len;
-			if (show_description && groupname_len > tinrc.groupname_max_length) {
-				/* no need to search further, we have reached max length */
-				groupname_len = tinrc.groupname_max_length;
-				break;
+	if (sel_fmt.len_grpname_max && !sel_fmt.len_grpname) {
+		/*
+		 * calculate max length of groupname field
+		 * if yanked in (yanked_out == FALSE) check all groups in active file
+		 * otherwise just subscribed to groups
+		 */
+		if (yanked_out) {
+			for (i = 0; i < selmenu.max; i++) {
+				if ((len = strwidth(active[my_group[i]].name)) > sel_fmt.len_grpname)
+					sel_fmt.len_grpname = len;
 			}
-		}
-	} else {
-		for_each_group(i) {
-			if ((len = strlen(active[i].name)) > groupname_len)
-				groupname_len = len;
-			if (show_description && groupname_len > tinrc.groupname_max_length) {
-				/* no need to search further, we have reached max length */
-				groupname_len = tinrc.groupname_max_length;
-				break;
+		} else {
+			for_each_group(i) {
+				if ((len = strwidth(active[i].name)) > sel_fmt.len_grpname)
+					sel_fmt.len_grpname = len;
 			}
 		}
 	}
-	if (groupname_len >= (cCOLS - SELECT_MISC_COLS))
-		groupname_len = cCOLS - SELECT_MISC_COLS - 1;
+
+	groupname_len = (sel_fmt.show_grpdesc && show_description) ? (int) sel_fmt.len_grpname_dsc : (int) sel_fmt.len_grpname;
+
+	if (groupname_len > (int) sel_fmt.len_grpname_max)
+		groupname_len = sel_fmt.len_grpname_max;
 	if (groupname_len < 0)
 		groupname_len = 0;
 
+	if (!sel_fmt.len_grpdesc)
+		sel_fmt.len_grpdesc = sel_fmt.len_grpname_max - groupname_len;
+	else {
+		if (sel_fmt.len_grpdesc > sel_fmt.len_grpname_max - groupname_len)
+			sel_fmt.len_grpdesc = sel_fmt.len_grpname_max - groupname_len;
+	}
+
+	flags_offset = sel_fmt.flags_offset + (sel_fmt.g_before_f ? groupname_len : 0) + (sel_fmt.d_before_f ? sel_fmt.len_grpdesc : 0);
+	ucnt_offset = sel_fmt.ucnt_offset + (sel_fmt.g_before_u ? groupname_len : 0) + (sel_fmt.d_before_u ? sel_fmt.len_grpdesc : 0);
+
 	for (i = selmenu.first; i < selmenu.first + NOTESLINES && i < selmenu.max; i++)
 		build_gline(i);
 
@@ -609,76 +653,186 @@ static void
 build_gline(
 	int i)
 {
-	char tmp[10];
-	char active_name[255];
-	char group_descript[255];
+	char *sptr, *fmt, *buf;
 	char subs;
-#ifdef USE_CURSES
-	char sptr[BUFSIZ];
+	int n;
+	size_t j;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	char *name_buf = NULL;
+	char *desc_buf = NULL;
+	wchar_t *active_name, *active_name2, *active_desc, *active_desc2;
 #else
-	char *sptr = screen[INDEX2SNUM(i)].col;
+	char *active_name, *active_name2;
+	size_t fill, len_start;
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+
+#ifdef USE_CURSES
+	/*
+	 * Allocate line buffer
+	 * make it the same size like in !USE_CURSES case to simplify the code
+	 */
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		sptr = my_malloc(cCOLS * MB_CUR_MAX + 2);
+#	else
+		sptr = my_malloc(cCOLS + 2);
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+#else
+	sptr = screen[INDEX2SNUM(i)].col;
 #endif /* USE_CURSES */
-	int n, blank_len;
-
-	blank_len = (MIN(cCOLS, (int) sizeof(group_descript)) - (groupname_len + SELECT_MISC_COLS)) + (show_description ? 2 : 4);
-
-	if (active[my_group[i]].inrange)
-		strcpy(tmp, "    #");
-	else if (active[my_group[i]].newsrc.num_unread) {
-		int getart_limit;
-		long num_unread;
-
-		getart_limit = cmdline.args & CMDLINE_GETART_LIMIT ? cmdline.getart_limit : tinrc.getart_limit;
-		num_unread = active[my_group[i]].newsrc.num_unread;
-		if (getart_limit > 0 && getart_limit < num_unread)
-			num_unread = getart_limit;
-		strcpy(tmp, tin_ltoa(num_unread, 5));
-	} else
-		strcpy(tmp, "     ");
 
+	sptr[0] = '\0';
+	fmt = sel_fmt.str;
 	n = my_group[i];
 
-	/*
-	 * Display a flag for this group if needed
-	 * . Bogus groups are dumped immediately
-	 * . Normal subscribed groups may be
-	 *   ' ' normal, 'X' not postable, 'M' moderated, '=' renamed
-	 * . Newgroups are 'N'
-	 * . Unsubscribed groups are 'u'
-	 */
-	if (active[n].bogus)					/* Group is not in active list */
-		subs = 'D';
-	else if (active[n].subscribed)			/* Important that this preceeds Newgroup */
-		subs = group_flag(active[n].moderated);
-	else
-		subs = ((active[n].newgroup) ? 'N' : 'u'); /* New (but unsubscribed) group or unsubscribed group */
+	if (tinrc.draw_arrow)
+		strcat(sptr, "  ");
 
-	strncpy(active_name, active[n].name, groupname_len);
-	active_name[groupname_len] = '\0';
+	for (; *fmt; fmt++) {
+		if (*fmt != '%') {
+			strncat(sptr, fmt, 1);
+			continue;
+		}
+		switch (*++fmt) {
+			case '\0':
+				break;
 
-	if (show_description) {
-		if (active[n].description) {
-			strncpy(group_descript, active[n].description, blank_len);
-			group_descript[blank_len] = '\0';
-			sprintf(sptr, "  %c %s %s  %-*.*s  %-*.*s%s",
-				 subs, tin_ltoa(i + 1, 4), tmp,
-				 groupname_len, groupname_len, active_name,
-				 blank_len, blank_len, group_descript, cCRLF);
-		} else
-			sprintf(sptr, "  %c %s %s  %-*.*s  %s",
-				 subs, tin_ltoa(i + 1, 4), tmp,
-				 (groupname_len + blank_len),
-				 (groupname_len + blank_len), active[n].name, cCRLF);
-	} else {
-		if (tinrc.draw_arrow)
-			sprintf(sptr, "  %c %s %s  %-*.*s%s", subs, tin_ltoa(i + 1, 4), tmp, groupname_len, groupname_len, active_name, cCRLF);
-		else
-			sprintf(sptr, "  %c %s %s  %-*.*s%*s%s", subs, tin_ltoa(i + 1, 4), tmp, groupname_len, groupname_len, active_name, blank_len, " ", cCRLF);
+			case '%':
+				strncat(sptr, fmt, 1);
+				break;
+
+			case 'd':
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+				if (show_description && active[n].description) {
+					active_desc = char2wchar_t(active[n].description);
+					if (active_desc) {
+						if ((active_desc2 = wcspart(active_desc, sel_fmt.len_grpdesc, TRUE)) != NULL) {
+							desc_buf = wchar_t2char(active_desc2);
+							free(active_desc);
+							free(active_desc2);
+						}
+					}
+					if (desc_buf) {
+						strcat(sptr, desc_buf);
+						FreeAndNull(desc_buf);
+					}
+				} else {
+					buf = sptr + strlen(sptr);
+					for (j = 0; j < sel_fmt.len_grpdesc; ++j)
+						*buf++ = ' ';
+					*buf = '\0';
+				}
+#else
+				if (show_description && active[n].description) {
+					len_start = strwidth(sptr);
+					strncat(sptr, active[n].description, sel_fmt.len_grpdesc);
+					fill = sel_fmt.len_grpdesc - (strwidth(sptr) - len_start);
+				} else
+					fill = sel_fmt.len_grpdesc;
+				buf = sptr + strlen(sptr);
+				for (j = 0; j < fill; ++j)
+					*buf++ = ' ';
+				*buf = '\0';
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+				break;
+
+			case 'f':
+				/*
+				 * Display a flag for this group if needed
+				 * . Bogus groups are dumped immediately
+				 * . Normal subscribed groups may be
+				 *   ' ' normal, 'X' not postable, 'M' moderated, '=' renamed
+				 * . Newgroups are 'N'
+				 * . Unsubscribed groups are 'u'
+				 *
+				 * TODO: make flags configurable via tinrc?
+				 */
+				if (active[n].bogus)					/* Group is not in active list */
+					subs = 'D';
+				else if (active[n].subscribed)			/* Important that this precedes Newgroup */
+					subs = group_flag(active[n].moderated);
+				else
+					subs = ((active[n].newgroup) ? 'N' : 'u'); /* New (but unsubscribed) group or unsubscribed group */
+				buf = sptr + strlen(sptr);
+				*buf++ = subs;
+				*buf = '\0';
+				break;
+
+			case 'G':
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+				if ((active_name = char2wchar_t(active[n].name)) == NULL) /* If char2wchar_t() fails try again after replacing unprintable characters */
+					active_name = char2wchar_t(convert_to_printable(active[n].name, FALSE));
+
+				if (active_name && tinrc.abbreviate_groupname) {
+					active_name2 = abbr_wcsgroupname(active_name, groupname_len);
+					free(active_name);
+				} else
+					active_name2 = active_name;
+
+				if (active_name2 && (active_name = wcspart(active_name2, groupname_len, TRUE)) != NULL) {
+					free(active_name2);
+					name_buf = wchar_t2char(active_name);
+					free(active_name);
+				}
+				if (name_buf) {
+					strcat(sptr, name_buf);
+					FreeAndNull(name_buf);
+				}
+#else
+				if (tinrc.abbreviate_groupname)
+					active_name = abbr_groupname(active[n].name, groupname_len);
+				else
+					active_name = my_strdup(active[n].name);
+
+				active_name2 = my_malloc(groupname_len + 1);
+				snprintf(active_name2, groupname_len + 1, "%-*s", groupname_len, active_name);
+				strcat(sptr, active_name2);
+				free(active_name);
+				free(active_name2);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+				break;
+
+			case 'n':
+				strcat(sptr, tin_ltoa(i + 1, sel_fmt.len_linenumber));
+				break;
+
+			case 'U':
+				if (active[my_group[i]].inrange) {
+					buf = sptr + strlen(sptr);
+					for (j = 1; j < sel_fmt.len_ucnt; ++j)
+						*buf++ = ' ';
+					*buf++ = tinrc.art_marked_inrange;
+					*buf = '\0';
+				} else if (active[my_group[i]].newsrc.num_unread) {
+					int getart_limit;
+					t_artnum num_unread;
+
+					getart_limit = cmdline.args & CMDLINE_GETART_LIMIT ? cmdline.getart_limit : tinrc.getart_limit;
+					num_unread = active[my_group[i]].newsrc.num_unread;
+					if (getart_limit > 0 && getart_limit < num_unread)
+						num_unread = getart_limit;
+					strcat(sptr, tin_ltoa(num_unread, sel_fmt.len_ucnt));
+				} else {
+					buf = sptr + strlen(sptr);
+					for (j = 0; j < sel_fmt.len_ucnt; ++j)
+						*buf++ = ' ';
+					*buf = '\0';
+				}
+				break;
+
+			default:
+				break;
+		}
 	}
+#ifndef USE_CURSES
 	if (tinrc.strip_blanks)
 		strcat(strip_line(sptr), cCRLF);
+#endif /* !USE_CURSES */
 
 	WriteLine(INDEX2LNUM(i), sptr);
+
+#ifdef USE_CURSES
+	free(sptr);
+#endif /* USE_CURSES */
 }
 
 
@@ -746,7 +900,7 @@ yank_active_file(
 
 
 /*
- * Sort active[] and associated qsort() helper function
+ * Sort active[] and associated tin_sort() helper function
  */
 static int
 active_comp(
@@ -809,7 +963,7 @@ sort_active_file(
 {
 	save_restore_curr_group(TRUE);
 
-	qsort(active, (size_t) num_active, sizeof(struct t_group), active_comp);
+	tin_sort(active, (size_t) num_active, sizeof(struct t_group), active_comp);
 	group_rehash(yanked_out);
 
 	selmenu.curr = save_restore_curr_group(FALSE);
@@ -917,7 +1071,7 @@ reposition_group(
 		return default_num;
 
 	if (strlen(pos))
-		pos_num = ((pos[0] == '$') ? selmenu.max: atoi(pos));
+		pos_num = ((pos[0] == '$') ? selmenu.max : atoi(pos));
 	else {
 		if (tinrc.default_move_group)
 			pos_num = tinrc.default_move_group;
@@ -972,7 +1126,15 @@ catchup_group(
 
 	if ((!TINRC_CONFIRM_ACTION) || prompt_yn(sized_message(&smsg, _(txt_mark_group_read), group->name), TRUE) == 1) {
 		grp_mark_read(group, NULL);
-		mark_screen(selmenu.curr, 9, "     ");
+		{
+			char buf[LEN];
+			size_t i = 0;
+
+			while (i < sel_fmt.len_ucnt)
+				buf[i++] = ' ';
+			buf[i] = '\0';
+			mark_screen(selmenu.curr, ucnt_offset, buf);
+		}
 
 		if (goto_next_unread_group)
 			pos_next_unread_group(TRUE);
@@ -1179,7 +1341,7 @@ subscribe_pattern(
 		return;
 	}
 
-	wait_message(0, message);
+	wait_message(0, "%s", message);
 
 	for_each_group(i) {
 		if (match_group_list(active[i].name, buf)) {
@@ -1255,3 +1417,264 @@ select_read_group(
 	else
 		info_message(_(txt_no_arts));
 }
+
+
+#ifdef NNTP_ABLE
+/*
+ * Try to fetch articles Nesgroups-header via [X]HDR or XPAT.
+ */
+static char *
+lookup_msgid(
+	char *msgid)
+{
+	if (read_news_via_nntp && !read_saved_news) {
+		if (!nntp_caps.hdr_cmd && !nntp_caps.xpat) {
+			info_message(_(txt_lookup_func_not_available));
+			return NULL;
+		}
+		if (msgid) {
+			char *ptr, *r = NULL;
+			static char *x = NULL;
+			char buf[NNTP_STRLEN];
+			int ret;
+
+			if (nntp_caps.hdr_cmd) {
+				snprintf(buf, sizeof(buf), "%s Newsgroups %s", nntp_caps.hdr_cmd, msgid);
+					ret = new_nntp_command(buf, (nntp_caps.type == CAPABILITIES) ? OK_HDR : OK_HEAD, NULL, 0);
+
+				switch (ret) {
+					case OK_HEAD:
+					case OK_HDR:
+						while ((ptr = tin_fgets(FAKE_NNTP_FP, FALSE)) != NULL) {
+#       ifdef DEBUG
+							if (debug & DEBUG_NNTP)
+								debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
+#       endif /* DEBUG */
+
+							if (ret == OK_HEAD) { /* RFC 2980 ("%s %s", id, grp) */
+								if (!strncmp(ptr, msgid, strlen(msgid))) { /* INN, MPNews, Leafnode, Cnews nntpd */
+									r = ptr + strlen(msgid) + 1;
+								} else { /* DNEWS ("%d %s", num, grp) */
+									r = ptr;
+									while (*r && *r != ' ' && *r != '\t')
+										r++;
+									while (*r && (*r == ' ' || *r == '\t'))
+										r++;
+								}
+							}
+
+							if (ret == OK_HDR) { /* RFC 3977 ("0 %s", grp) */
+								if (*ptr == '0' && (*(ptr + 1) == ' ' || *(ptr + 1) == '\t'))
+									r = ptr + 2;
+
+							}
+
+							if (r)
+								x = my_strdup(r);
+						}
+
+						if (x)
+							return x;
+
+						if (!r) {
+#       ifdef DEBUG
+								if (debug & DEBUG_NNTP)
+									debug_print_file("NNTP", "lookup_msgid(%s) response empty or not recognized", buf);
+#       endif /* DEBUG */
+								if (!nntp_caps.xpat)
+									info_message(_(txt_lookup_func_not_available));
+						}
+						if (r || !nntp_caps.xpat)
+							return NULL;
+						break;
+
+					case ERR_NOART:
+						info_message(_(txt_art_unavailable));
+						return NULL;
+
+					default:
+						if (!nntp_caps.xpat) { /* try only once */
+							info_message(_(txt_lookup_func_not_available));
+							return NULL;
+						}
+						break;
+				}
+			}
+
+			if (nntp_caps.xpat) {
+				snprintf(buf, sizeof(buf), "XPAT Newsgroups %s *", msgid);
+				ret = new_nntp_command(buf, OK_HEAD, NULL, 0);
+				r = NULL;
+				switch (ret) {
+					case OK_HEAD:
+						while ((ptr = tin_fgets(FAKE_NNTP_FP, FALSE)) != NULL) {
+#ifdef DEBUG
+							if (debug & DEBUG_NNTP)
+								debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
+#endif /* DEBUG */
+							if (!strncmp(ptr, msgid, strlen(msgid)))
+								r = ptr + strlen(msgid) + 1;
+
+							if (r)
+								x = my_strdup(r);
+						}
+
+						if (x)
+							return x;
+
+						if (!r) {
+#       ifdef DEBUG
+								if (debug & DEBUG_NNTP)
+									debug_print_file("NNTP", "lookup_msgid(%s) response empty or not recognized", buf);
+#       endif /* DEBUG */
+								info_message(_(txt_lookup_func_not_available));
+								/* nntp_caps.xpat = FALSE; */ /* ? */
+						}
+						return NULL;
+
+					case ERR_NOART:
+						info_message(_(txt_art_unavailable));
+						return NULL;
+
+					default:
+						nntp_caps.xpat = FALSE;
+						break;
+				}
+			}
+			info_message(_(txt_lookup_func_not_available));
+		}
+	} else
+		info_message("%s %s", _(txt_lookup_func_not_available), _(txt_lookup_func_not_nntp));
+
+	return NULL;
+}
+
+
+/*
+ * Get a message ID for the 'L' command. Add <> if needed.
+ * Try to enter an appropriate group and display the referenced article.
+ * If no group from the Newsgroups:-header is available, display the
+ * contents of the header.
+ */
+static int
+show_article_by_msgid(
+	void)
+{
+	char id[LEN];
+	char *idptr;
+	char *newsgroups = NULL;
+	int i, ret = 0;
+	struct t_article *art;
+	struct t_group *group = NULL;
+	struct t_msgid *msgid = NULL;
+	t_bool tmp_cache_overview_files;
+	t_bool tmp_show_only_unread_arts;
+
+	if (!(read_news_via_nntp && !read_saved_news)) {
+		info_message("%s %s", _(txt_lookup_func_not_available), _(txt_lookup_func_not_nntp));
+		return -1;
+	}
+
+	if (prompt_string(_(txt_enter_message_id), id + 1, HIST_MESSAGE_ID) && id[1]) {
+		idptr = str_trim(id + 1);
+		if (id[1] != '<') {
+			id[0] = '<';
+			strcat(id, ">");
+			idptr = id;
+		}
+		newsgroups = lookup_msgid(idptr);
+	}
+
+	if (!newsgroups)
+		return -1;
+
+	if ((group = get_group_from_list(newsgroups)) == NULL) {
+		info_message(strchr(newsgroups, ',') ? _(txt_lookup_show_groups) : _(txt_lookup_show_group), newsgroups);
+		free(newsgroups);
+		return -1;
+	}
+
+	curr_group = group;
+	num_of_tagged_arts = 0;
+	range_active = FALSE;
+	last_resp = -1;
+	this_resp = -1;
+	tmp_cache_overview_files = tinrc.cache_overview_files;
+	tinrc.cache_overview_files = FALSE;
+	tmp_show_only_unread_arts = curr_group->attribute->show_only_unread_arts;
+	curr_group->attribute->show_only_unread_arts = FALSE;
+
+	if (!index_group(group)) {
+		for_each_art(i) {
+			art = &arts[i];
+			FreeAndNull(art->refs);
+			FreeAndNull(art->msgid);
+		}
+		tin_errno = 0;
+		ret = -1;
+	}
+
+	if (!ret) {
+		grpmenu.first = 0;
+
+		if ((msgid = find_msgid(idptr)) == NULL) {
+			info_message(_(txt_art_unavailable));
+			ret = -1;
+		}
+
+		if (!ret && msgid->article == ART_UNAVAILABLE) {
+			info_message(_(txt_art_unavailable));
+			ret = -1;
+		}
+
+		if (!ret && which_thread(msgid->article) == -1) {
+			info_message(_(txt_no_last_message));
+			ret = -1;
+		}
+	}
+
+	if (!ret) {
+		switch ((i = show_page(group, msgid->article, NULL))) {
+			case GRP_QUIT:
+				ret = GRP_QUIT;
+				break;
+
+			default:
+				break;
+		}
+	}
+
+	free(newsgroups);
+	art_close(&pgart);
+	tinrc.cache_overview_files = tmp_cache_overview_files;
+	curr_group->attribute->show_only_unread_arts = tmp_show_only_unread_arts;
+	curr_group = NULL;
+
+	return ret;
+}
+
+
+/*
+ * Takes a list of newsgroups and determines if one of them is available.
+ */
+static struct t_group *
+get_group_from_list(
+	char *newsgroups)
+{
+	char *ptr;
+	t_bool found = FALSE;
+	struct t_group *group = NULL;
+
+	if (!newsgroups || (ptr = strtok(newsgroups, ",")) == NULL)
+		return NULL;
+
+	/* find first available group of type news */
+	do {
+		group = group_find(ptr, TRUE);
+		if (group && group->type == GROUP_TYPE_NEWS)
+			found = TRUE;
+	} while (!found && (ptr = strtok(NULL, ",")) != NULL);
+
+	return found ? group : NULL;
+}
+#endif /* NNTP_ABLE */
